% code to solve nonisotermal chemical reactor with axial dispersion
clear all;
addpath('mn');       % location of function files .m and .mex
global fout symm geom meth Neq n2 Erx Pe Bc0 Bc1 Rx Bi nv
run OCCdefs                % script file with useful definitions
symm = NonSymmetric;       % symmetry
meth = Lobatto;            % point type
geom = Planar;             % geometry
Neq = 2;                   % no. coupled equations
Pe = [1.0/100.,1.0/200.];  % Peclet no. (inverse)
Bi = [3.0,0.0];            % Biot no. (0 for mass)
Rx = [0.2,0.5];            % Damkohler no. (rate coefficients)
Bc0 = 1.0 + Bi.*Pe;        % boundary conditions (inlet)
Bc1 = Bi.*Pe;              % boundary conditions (outlet)
Korder = 1;                % reaction order, not implemented
Erx = 20.0;                % activation energy term
BoundaryColloc = 0;        % 0 natural bc, 1 boundary collocation
ncalc = [11,11,2];
fout = fopen('ex3x.dat','w');
maxiter = 100;  tol = 1.0e-6; %eps*1.0e5;
ch = input ( 'Enter number interior points first, last, delta >','s' );
ncalc = sscanf(ch,'%i');
ch = input( 'Enter solution method 0/1/2 >','s' );
MethSolv = str2num(ch);
% ---------------------------------------------------------
function [r,dr] = RxnRate(y)   % reaction rate w/derivatives
   global fout symm geom meth Neq n2 Erx
   Tinv(:,1) = (1.0)./(y(:,1) + 1.0);
   dr(:,2) = -exp(Erx*(1.0 - Tinv));  % 1st order nonisothermal
   r(:,1) = -dr(:,2).*(1.0 - y(:,2));
   dr(:,1) = Erx.*Tinv.*Tinv.*r;
end % -----------------------------------------------------
function [rsd,rbc] = Residual(Bc,Ac,wc,Lc,y,rv)    % Calculate Residuals
   global fout symm geom meth Neq n2 Erx Pe Bc0 Bc1 Rx Bi nv
   rsd(1:nv,1:Neq) = 0.0; rbc(1:2,1:Neq) = 0.0;
   for j=1:Neq
      rd = (Pe(j).*Bc - Ac)*y(:,j) - Bi(j)*y(:,j); % polynomial, linear terms
      rsd(:,j) = Lc*rd + Rx(j)*rv;                 % residual
      rbc(1,j) = Bc0(j)*y(1,j) - Pe(j)*Ac(1,:)*y(:,j);
      rbc(2,j) = Bc1(j)*y(n2,j) + Pe(j)*Ac(n2,:)*y(:,j);
   end
end % -----------------------------------------------------
function [G] = CalcCoef(Cc,Ac,wc)  % calc. fixed coefficients
   global fout symm geom meth Neq n2 Erx Pe Bc0 Bc1 Rx Bi
   Aw = bsxfun(@times,wc,Ac);     % W*A
   for j=1:Neq
      G(:,:,j) = Pe(j).*Cc .+ Aw + diag(wc(:)*Bi(j));
      G(1,1,j) = G(1,1,j) + Bc0(j);
      G(n2,n2,j) = G(n2,n2,j) + Bc1(j);
   end
end % -----------------------------------------------------
function dy = CalcReduce(G,wc,y,r,dr) % Solution using reduced NR method
   global fout symm geom meth Neq n2 Erx Pe Bc0 Bc1 Rx Bi
   b(:,1) = Rx(2)*r.*wc - G(:,:,2)*y(:,2);
   GG = G(:,:,2) - bsxfun(@times,Rx(1)*wc.*dr(:,1),G(:,:,1)) - diag(Rx(2)*wc.*dr(:,2));
   dy(:,2) = GG\b;
   dy(:,1) = (Rx(1)/Rx(2))*G(:,:,1)*dy(:,2);
end % -----------------------------------------------------
function dy = CalcFull(G,wc,y,r,dr) % Solution full NR matrix problem
   global fout symm geom meth Neq n2 Erx Pe Bc0 Bc1 Rx Bi
   GG(1:n2*Neq,1:n2*Neq) = 0.0;  bg(1:n2*Neq) = 0.0;
   for i=1:Neq  % load coefficients into full matrix
      i1 = 1 + (i-1)*n2;  i2 = i*n2;
      bg(i1:i2) = Rx(i)*r.*wc - G(:,:,i)*y(:,i);   % residual
      for j=1:Neq
         j1 = 1 + (j-1)*n2;  j2 = j*n2;
         if(i == j)
            GG(i1:i2,j1:j2) = G(:,:,i) - diag(Rx(i)*dr(:,j).*wc);
         else
            GG(i1:i2,j1:j2) = diag(-Rx(i)*dr(:,j).*wc);
         end
      end
   end
   dy = reshape(GG\bg',n2,Neq);
end % -----------------------------------------------------
function [dy] = CalcPicard(G,L,U,wc,y,r,MethSolv) % Solution w/ Picard iteration
   global fout symm geom meth Neq n2 Erx Pe Bc0 Bc1 Rx Bi
      for j=1:Neq
      b(:,j) = Rx(j)*r.*wc - G(:,:,j)*y(:,j);
%     dy(:,j) = G(:,:,j)\b(:,j);    % factors every iteration
      dy(:,j) = U(:,:,j)\(L(:,:,j)\b(:,j));  % should be more efficient
   end
end % -----------------------------------------------------
for n=ncalc(1):ncalc(3):ncalc(2)  % loop over number of interior points
   n2 = n + 2;
   fprintf(fout,'Solution with %s points, n =%3i\n',strtrim(strmeth(meth,:)),n);
   fprintf(1,'Solution with %s points, n =%3i\n',strtrim(strmeth(meth,:)),n);
   [xc,wc,Ac] = OCnonsym(n,meth);       % points, quadrature wts., 1st derivative
   [Bc,Cc] = OCBCcoef(wc,Ac,Ac,symm);   % Laplacian and stiffness
   if(BoundaryColloc)
      disp('Using Boundary Collocation');
      wc(1) = 0.0;  wc(n2) = 0.0;
      Cc(1,:) = -Ac(1,:);   Cc(n2,:) = Ac(n2,:);
   end
   y(1:n2,1:Neq) = 0.0;  dy(1:n2,1:Neq) = 0.0;  % initialize
   G = CalcCoef(Cc,Ac,wc);
   switch MethSolv  % Setup for iterative methods
   case 0
      G(:,:,1) = G(:,:,1)\G(:,:,2);
      fprintf(fout,'Reduced N-R Iteration\n');
      fprintf(1,'Reduced N-R Iteration\n');
   case 1
      fprintf(fout,'Full N-R Iteration\n');
      fprintf(1,'Full N-R Iteration\n');
   otherwise
      Piter = [0.25,+0.25];  % iteration parameters
      fprintf(fout,'Picard Iteration Parms =  %8.4f %8.4f\n',Piter(1),Piter(2));
      fprintf(1,'Picard Iteration Parms =  %8.4f %8.4f\n',Piter(1),Piter(2));
      L(1:n2,1:n2,1:Neq) = 0.0;  U(1:n2,1:n2,1:Neq) = 0.0;
      for j=1:Neq
         dx = Rx(j)*wc*Piter(min(j,2)); % scaled iteration parms.
         [L(:,:,j),U(:,:,j)] = lu(G(:,:,j)+diag(dx));
      end
   end
   for k=1:maxiter % Iteration Loop -------------
      [r,dr] = RxnRate(y);
      switch MethSolv
      case 0
         dy = CalcReduce(G,wc,y,r,dr,k);  % reduced n2*(Neq-1)
      case 1
         dy = CalcFull(G,wc,y,r,dr);      % full n2*Neq
      otherwise
         dy = CalcPicard(G,L,U,wc,y,r,MethSolv);   % Picard iteration
      end
      y = y + dy;
      rnorm = max(max(abs(dy))); % max rows and cols.
      if(k < 10)
         ArrayPrint(fout,y','Solution New');
      end
      fprintf(fout,'Iteration %2i, Norm = %12.4e\n',k,rnorm);
      fprintf(1,'Iteration %2i, Norm = %12.4e\n',k,rnorm);
      if(rnorm < tol)break; end
   end % ----------------------------------------
   nf = 4; nv = 1 + (n+1)*nf;  % create dense x, fill between xc
   xv(1:nv,1) = 0.0; yv(1:nv,1:Neq) = 0.0; rsd(1:nv,1:Neq+1) = 0.0;
   dx = xc(2:n2) - xc(1:n2-1);
   for i=1:nf
      wt = double(i)/double(nf);
      xv(i+1:nf:nv) = xc(1:n+1) + dx*wt;
   end
   Lc = Lcoef(xv,xc);
   for i=1:Neq
      yv(:,i) = Lc*y(:,i);    % interpolate solution
   end
   ArrayPrintX(fout,xc,y,r,'Results: x, y1, y2, r');
   rsd(:,1) = RxnRate(yv);
   [rsd(:,2:Neq+1),rbc] = Residual(Bc,Ac,wc,Lc,y,rsd(:,1));
   fprintf(fout,'\nInlet Resid. :\t%11.5e\t%11.5e\t%11.5e\t%11.5e\n',rbc(1,1),rsd(1,2),rbc(1,2),rsd(1,3));
   fprintf(fout,'Outlet Resid.:\t%11.5e\t%11.5e\t%11.5e\t%11.5e\n',rbc(2,1),rsd(nv,2),rbc(2,2),rsd(nv,3));
   ArrayPrintX(fout,xv,yv,rsd,'Results: x, y1, y2, r, rsd1, rsd2');
end
fclose(fout);

